@@ -6,10 +6,14 @@ import platform
 import sys
 import logging
 import asyncio
+from typing import Optional, Union
 import inspect
+import hashlib
+import hmac
 
 # ThirdParty Imports
 import aiohttp
+from aiohttp import FormData
 
 # Internal Imports
 from slack.web.slack_response import SlackResponse
@@ -25,11 +29,12 @@ class BaseClient:
         token,
         base_url=BASE_URL,
         timeout=30,
-        loop=None,
+        loop: Optional[asyncio.AbstractEventLoop] = None,
         ssl=None,
         proxy=None,
         run_async=False,
         session=None,
+        headers: Optional[dict] = None,
     ):
         self.token = token
         self.base_url = base_url
@@ -38,16 +43,44 @@ class BaseClient:
         self.proxy = proxy
         self.run_async = run_async
         self.session = session
+        self.headers = headers or {}
         self._logger = logging.getLogger(__name__)
         self._event_loop = loop
 
-    def _set_event_loop(self):
-        if self.run_async:
-            self._event_loop = asyncio.get_event_loop()
-        else:
+    def _get_event_loop(self):
+        """Retrieves the event loop or creates a new one."""
+        try:
+            return asyncio.get_event_loop()
+        except RuntimeError:
             loop = asyncio.new_event_loop()
             asyncio.set_event_loop(loop)
-            self._event_loop = loop
+            return loop
+
+    def _get_headers(self, has_json, has_files):
+        """Contructs the headers need for a request.
+        Args:
+            has_json (bool): Whether or not the request has json.
+        Returns:
+            The headers dictionary.
+                e.g. {
+                    'Content-Type': 'application/json;charset=utf-8',
+                    'Authorization': 'Bearer xoxb-1234-1243',
+                    'User-Agent': 'Python/3.6.8 slack/2.1.0 Darwin/17.7.0'
+                }
+        """
+        headers = {
+            "User-Agent": self._get_user_agent(),
+            "Authorization": "Bearer {}".format(self.token),
+            "Content-Type": "application/x-www-form-urlencoded;charset=utf-8",
+        }
+        if has_json:
+            headers.update({"Content-Type": "application/json;charset=utf-8"})
+
+        if has_files:
+            # These are set automatically by the aiohttp library.
+            headers.pop("Content-Type", None)
+
+        return headers
 
     def api_call(
         self,
@@ -55,10 +88,10 @@ class BaseClient:
         *,
         http_verb: str = "POST",
         files: dict = None,
-        data: dict = None,
+        data: Union[dict, FormData] = None,
         params: dict = None,
         json: dict = None,
-    ):
+    ) -> Union[asyncio.Future, SlackResponse]:
         """Create a request and execute the API call to Slack.
 
         Args:
@@ -89,32 +122,18 @@ class BaseClient:
             SlackRequestError: Json data can only be submitted as
                 POST requests.
         """
-        if json is not None and http_verb != "POST":
+        has_json = json is not None
+        has_files = files is not None
+        if has_json and http_verb != "POST":
             msg = "Json data can only be submitted as POST requests. GET requests should use the 'params' argument."
             raise err.SlackRequestError(msg)
 
         api_url = self._get_url(api_method)
-        headers = {
-            "User-Agent": self._get_user_agent(),
-            "Authorization": "Bearer {}".format(self.token),
-        }
-        if files is not None:
-            form_data = aiohttp.FormData()
-            for k, v in files.items():
-                if isinstance(v, str):
-                    form_data.add_field(k, open(v, "rb"))
-                else:
-                    form_data.add_field(k, v)
-
-            if data is not None:
-                for k, v in data.items():
-                    form_data.add_field(k, str(v))
-
-            data = form_data
 
         req_args = {
-            "headers": headers,
+            "headers": self._get_headers(has_json, has_files),
             "data": data,
+            "files": files,
             "params": params,
             "json": json,
             "ssl": self.ssl,
@@ -122,14 +141,14 @@ class BaseClient:
         }
 
         if self._event_loop is None:
-            self._set_event_loop()
+            self._event_loop = self._get_event_loop()
 
         future = asyncio.ensure_future(
             self._send(http_verb=http_verb, api_url=api_url, req_args=req_args),
             loop=self._event_loop,
         )
 
-        if self.run_async or self._event_loop.is_running():
+        if self.run_async:
             return future
 
         return self._event_loop.run_until_complete(future)
@@ -177,9 +196,24 @@ class BaseClient:
         Returns:
             The response parsed into a SlackResponse object.
         """
+        open_files = []
+        files = req_args.pop("files", None)
+        if files is not None:
+            for k, v in files.items():
+                if isinstance(v, str):
+                    f = open(v.encode("ascii", "ignore"), "rb")
+                    open_files.append(f)
+                    req_args["data"].update({k: f})
+                else:
+                    req_args["data"].update({k: v})
+
         res = await self._request(
             http_verb=http_verb, api_url=api_url, req_args=req_args
         )
+
+        for f in open_files:
+            f.close()
+
         data = {
             "client": self,
             "http_verb": http_verb,
@@ -195,7 +229,6 @@ class BaseClient:
         """
         if self.session and not self.session.closed:
             async with self.session.request(http_verb, api_url, **req_args) as res:
-                self._logger.debug("Ran the request with existing session.")
                 return {
                     "data": await res.json(),
                     "headers": res.headers,
@@ -205,7 +238,6 @@ class BaseClient:
             loop=self._event_loop, timeout=aiohttp.ClientTimeout(total=self.timeout)
         ) as session:
             async with session.request(http_verb, api_url, **req_args) as res:
-                self._logger.debug("Ran the request with a new session.")
                 return {
                     "data": await res.json(),
                     "headers": res.headers,
@@ -219,13 +251,45 @@ class BaseClient:
 
         Returns:
             The user agent string.
-            e.g. 'Python/3.6.7 slack/2.0.0 Darwin/17.7.0'
+            e.g. 'Python/3.6.7 slackclient/2.0.0 Darwin/17.7.0'
         """
         # __name__ returns all classes, we only want the client
-        client = "{0}/{1}".format(__name__.split(".")[0], ver.__version__)
+        client = "{0}/{1}".format("slackclient", ver.__version__)
         python_version = "Python/{v.major}.{v.minor}.{v.micro}".format(
             v=sys.version_info
         )
         system_info = "{0}/{1}".format(platform.system(), platform.release())
         user_agent_string = " ".join([python_version, client, system_info])
         return user_agent_string
+
+    @staticmethod
+    def validate_slack_signature(
+        *, signing_secret: str, data: str, timestamp: str, signature: str
+    ) -> bool:
+        """
+        Slack creates a unique string for your app and shares it with you. Verify
+        requests from Slack with confidence by verifying signatures using your
+        signing secret.
+
+        On each HTTP request that Slack sends, we add an X-Slack-Signature HTTP
+        header. The signature is created by combining the signing secret with the
+        body of the request we're sending using a standard HMAC-SHA256 keyed hash.
+
+        https://api.slack.com/docs/verifying-requests-from-slack#how_to_make_a_request_signature_in_4_easy_steps__an_overview
+
+        Args:
+            signing_secret: Your application's signing secret, available in the
+                Slack API dashboard
+            data: The raw body of the incoming request - no headers, just the body.
+            timestamp: from the 'X-Slack-Request-Timestamp' header
+            signature: from the 'X-Slack-Signature' header - the calculated signature
+                should match this.
+
+        Returns:
+            True if signatures matches
+        """
+        format_req = str.encode(f"v0:{timestamp}:{data}")
+        encoded_secret = str.encode(signing_secret)
+        request_hash = hmac.new(encoded_secret, format_req, hashlib.sha256).hexdigest()
+        calculated_signature = f"v0={request_hash}"
+        return hmac.compare_digest(calculated_signature, signature)
